(* Module Main: The main program.  Deals with processing the command
   line, reading files, building and connecting lexers and parsers, etc. 
   
   For most experiments with the implementation, it should not be
   necessary to change this file.
*)

open Format
open Support.Pervasive
open Support.Error
open Syntax
open Core

let searchpath = ref [""]
let ignoreErrors = ref false

let argDefs = [
  "-I",
      Arg.String (fun f -> searchpath := f::!searchpath),
      "Append a directory to the search path";
  "-s",
      Arg.Bool (fun b -> ignoreErrors := b),
      "Skip errors and continue executing"]

let parseArgs () =
  let inFile = ref (None : string option) in
  Arg.parse argDefs
     (fun s ->
       match !inFile with
         Some(_) -> err "You must specify exactly one input file"
       | None -> inFile := Some(s))
     "";
  match !inFile with
      None -> err "You must specify an input file"
    | Some(s) -> s

let openfile infile = 
  let rec trynext l = match l with
        [] -> err ("Could not find " ^ infile)
      | (d::rest) -> 
          let name = if d = "" then infile else (d ^ "/" ^ infile) in
          try open_in name
            with Sys_error m -> trynext rest
  in trynext !searchpath

let parseFile inFile =
  let pi = openfile inFile
  in let lexbuf = Lexer.create inFile pi
  in let result =
    try Parser.toplevel Lexer.main lexbuf with Parsing.Parse_error -> 
    error (Lexer.info lexbuf) "Parse error"
in
  Parsing.clear_parser(); close_in pi; result

let alreadyImported = ref ([] : string list)

let checkprog prog fundefs = 
  let break() = print_break 0 0 in
  let check def = match def with
	(_,FnDef(_,_,_,None)) -> ()
      | (fn,FnDef(fi,args,resty,Some bdy)) -> 
	  let binds = List.map (fun (x,ty) -> (x,VarBind ty)) args in
	  let ctx = List.fold_left (addbinding fi) emptycontext binds in
	  let ty = typeof prog ctx bdy in
	  if (=) ty resty then ()
	  else errfAt fi (fun()->
			   pr ("Body of function \"" ^ fn ^ "\" has type ");
	                   printty ty;
			   pr " and not type ";
			   printty resty)
  in List.iter check fundefs

let rec process_command (prog,ctx) cmd = match cmd with
  | Eval(fi,t) ->
      let tyT = typeof prog ctx t in
      let t' = eval prog t in
      printtm t'; 
      print_break 1 2;
      pr ": ";
      printty tyT;
      force_newline();
      (prog,ctx)
  | UnsafeEval(fi,t) -> 
      let t' = eval prog t in
      printtm t'; 
      force_newline();
      (prog,ctx)
  | Bind(fi,x,bind) -> 
      pr x; pr " "; prbinding bind; force_newline();
      (prog,addbinding fi ctx (x,bind))
  | Prog(fi,fundefs) -> 
      let prog = makeprog (List.rev fundefs) in
      (checkprog prog fundefs;
       prprog prog; force_newline();
       (prog, emptycontext))
  
let process_file f prog ctx  =
  alreadyImported := f :: !alreadyImported;
  let cmds = parseFile f in
  let g  c =  
    open_hvbox 0;
    let results = 
      if !ignoreErrors then 
	(fun cmd -> try process_command c cmd
		    with Exit _ -> c)
      else process_command c
    in 
    print_flush();
    results
  in
    List.fold_left g (prog,ctx) cmds

let main () = 
  let inFile = parseArgs() in
  let _ = process_file inFile emptyprog emptycontext in
  ()

let () = set_max_boxes 1000
let () = set_margin 67
let res = 
  Printexc.catch (fun () -> 
    try main();0 
    with Exit x -> x) 
  ()
let () = print_flush()
let () = exit res
